//  aplist.h
//
//  DRAFT
//
//  Chris Nevison, Mark Weiss
//  February 2, 1998
//
/////////////////////////////////////////////////////////////////////////
//  An implementation of a two-way list with iterator, a container
//  class, with ItemType specifying the type of item on the list. 
//
//  The template class aplist defines a two-way list structure.
//
//  The template class aplistIterator defines a two-way iterator class
//  used to traverse through the list and access items in the list.
//
//  The type ItemType specifies the type of items on the list,
//  which need to have the following member functions defined:
//     zero-parameter (default) constructor
//     operator = and copy constructor
//
//  The two classes aplist and aplistIterator must be declared as 
//  mutual friends.
/////////////////////////////////////////////////////////////////////////

template <class ItemType>     // Forward reference so aplistIterator can be
class aplistIterator;         // declared as friend of aplist and aplistNode

template <class ItemType>     // Forward reference so aplist can be
class aplist;                 // declared as friend of aplistNode


template <class ItemType>
class aplistNode{ 
public:
  friend class aplist<ItemType>;
  friend class aplistIterator<ItemType>;

private:
  aplistNode(const ItemType & item, aplistNode * prev, aplistNode * next);

  ItemType     myItem;         // All members are private because          
  aplistNode * myPrev;         // aplistNode is not intended for client
  aplistNode * myNext;         // use, but only for implementation.
};                             // Consequently aplist and aplistIterator
                               // are declared as friend classes.

template <class ItemType>
class aplist
{
public:

  aplist();                             // default constructor
  aplist(const aplist<ItemType> & rhs); // copy constructor
  ~aplist();                            // destructor

  const aplist<ItemType> & 
    operator = (const aplist<ItemType> & rhs); // assignment operator

  //  Accessor

  bool  IsEmpty()  const;      // returns true if the list is empty

  aplistIterator<ItemType>
        First()    const;      // returns iterator referencing the first
                               // item in the list (EndMarker if list is empty)
  aplistIterator<ItemType>
        Last()     const;      // returns iterator referencing the last
                               // item in the list (FrontMarker if list is empty)
  aplistIterator<ItemType>
        FrontMarker() const;   // returns iterator referencing before the 
                               // first item in the list
  aplistIterator<ItemType>
        EndMarker() const;     // returns iterator referencing past the 
                               // last item in the list
  
  //  Modifiers

  void  MakeEmpty();                      // Make the list empty, all 
                                          // iterators become invalid
  void  Append(const ItemType & item);    // Add item at end of list
  void  Prepend(const ItemType & item);   // Add item at front of list

  aplistIterator<ItemType> 
  InsertBefore(const aplistIterator<ItemType> & itr,// Insert item into the list
               const ItemType & item);              // before position itr, returns
                                                    // an iterator referencing item
  aplistIterator<ItemType> 
  InsertAfter(const aplistIterator<ItemType> & itr, // Insert item into the list
              const ItemType & item);               // after position itr, returns
                                                    // an iterator referencing item

  void Erase(aplistIterator<ItemType> & itr); // Removes the item referenced by
                                              // itr, changes itr to reference the
                                              // next item in the list
           // Note:  The member functions Append, Prepend, Insert
           //        do not change exiting iterators.
           //        Erase changes only the iterator passed as a
           //        a parameter and any other iterators referencing
           //        the same item become invalid.
           //        MakeEmpty invalidates all iterators except those
           //        which reference the FrontMarker or EndMarker positions
  
  friend class aplistIterator<ItemType>;    // permits aplistIterator to
                                            // access private members of aplist
private:

  aplistNode<ItemType> * myHead;
  aplistNode<ItemType> * myTail;

  // Helper function to create iterators

  aplistIterator<ItemType> 
  NewIterator(aplistNode<ItemType> * position) const;
};


template <class ItemType>
class aplistIterator
{
public:
  
  aplistIterator(const aplist<ItemType> & list); // default constructor
  aplistIterator(const aplistIterator<ItemType> & rhs); // copy constructor
  ~aplistIterator();                             // destructor
                   // Note: The aplist associated with the iterator
                   //       is a const reference parameter.  This
                   //       means the structure of the list can not be
                   //       changed by an iterator.  However the value of
                   //       an item referenced by the iterator can be
                   //       changed using the member function Current

  const aplistIterator & 
    operator = (const aplistIterator<ItemType> & rhs); // assignment

  //  Accessors

  bool             OnList()  const;  // Returns true if the iterator
                                     // references an item on the list
  const ItemType & Current() const;  // Returns a reference to the item
                                     // currently referenced by the iterator

  void   SetCurrent(const ItemType & item);  
                                     // Sets the current position  to the
                                     // contain the value item

  bool operator == (const aplistIterator<ItemType> & itr) const;
  bool operator != (const aplistIterator<ItemType> & itr) const;

  //   Modifiers, change the position of the iterator on the list

  void  Advance();          // Move to next item in list, or off list
  void  Backup();           // Move to previous item in list, or off list

  friend class aplist<ItemType>; // aplist must access iterators for 
                                 // Insert and Erase
private:

  const aplist<ItemType>  & myList;
  
  aplistNode<ItemType> * myPos;
};

#include "aplist.cpp"
